/*
 * This file was generated by qdbusxml2cpp version 0.8
 * Command line was: qdbusxml2cpp -c MainWindowAdaptor -a MainWindowAdaptor.h:MainWindowAdaptor.cpp
 * org.keepassxc.MainWindow.xml
 *
 * qdbusxml2cpp is Copyright (C) 2015 The Qt Company Ltd.
 *
 * This is an auto-generated file.
 * Do not edit! All changes made to it will be lost.
 */

#include "MainWindowAdaptor.h"
#include <QtCore/QByteArray>
#include <QtCore/QList>
#include <QtCore/QMap>
#include <QtCore/QMetaObject>
#include <QtCore/QString>
#include <QtCore/QStringList>
#include <QtCore/QVariant>

/*
 * Implementation of adaptor class MainWindowAdaptor
 */

MainWindowAdaptor::MainWindowAdaptor(QObject* parent)
    : QDBusAbstractAdaptor(parent)
{
    setAutoRelaySignals(true);
}

MainWindowAdaptor::~MainWindowAdaptor()
{
}

void MainWindowAdaptor::appExit()
{
    QMetaObject::invokeMethod(parent(), "appExit");
}

void MainWindowAdaptor::closeAllDatabases()
{
    QMetaObject::invokeMethod(parent(), "closeAllDatabases");
}

void MainWindowAdaptor::lockAllDatabases()
{
    QMetaObject::invokeMethod(parent(), "lockAllDatabases");
}

void MainWindowAdaptor::openDatabase(const QString& fileName)
{
    QMetaObject::invokeMethod(parent(), "openDatabase", Q_ARG(QString, fileName));
}

void MainWindowAdaptor::openDatabase(const QString& fileName, const QString& pw)
{
    QMetaObject::invokeMethod(parent(), "openDatabase", Q_ARG(QString, fileName), Q_ARG(QString, pw));
}

void MainWindowAdaptor::openDatabase(const QString& fileName, const QString& pw, const QString& keyFile)
{
    QMetaObject::invokeMethod(
        parent(), "openDatabase", Q_ARG(QString, fileName), Q_ARG(QString, pw), Q_ARG(QString, keyFile));
}
