#include "VersionInfo.h"
#include "winresrc.h"

VS_VERSION_INFO VERSIONINFO
    FILEVERSION FILE_VERSION_RESOURCE
    PRODUCTVERSION PRODUCT_VERSION_RESOURCE
    FILEFLAGSMASK 0x3fL
#ifdef _DEBUG
    FILEFLAGS 0x1L
#else
    FILEFLAGS 0x0L
#endif
    FILEOS 0x4L
    FILETYPE 0x1L
    FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904b0"
        BEGIN
            VALUE "Comments", PRODUCT_COMMENTS
            VALUE "CompanyName", PRODUCT_COMPANY_NAME
            VALUE "FileDescription", PRODUCT_FILE_DESCRIPTION
            VALUE "FileVersion", FILE_VERSION_RESOURCE_STR
            VALUE "InternalName", PRODUCT_INTERNAL_NAME
            VALUE "LegalCopyright", PRODUCT_COMPANY_COPYRIGHT
            VALUE "OriginalFilename", PRODUCT_ORIGINAL_FILENAME
            VALUE "ProductName", PRODUCT_BUNDLE
            VALUE "ProductVersion", PRODUCT_VERSION_RESOURCE_STR
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1200
    END
END
