/*
 *  Copyright (C) 2012 Felix Geyer <debfx@fobos.de>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 or (at your option)
 *  version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef KEEPASSX_WILDCARDMATCHER_H
#define KEEPASSX_WILDCARDMATCHER_H

#include <QStringList>

class WildcardMatcher
{
public:
    explicit WildcardMatcher(QString text);
    bool match(const QString& pattern);

    static const QChar Wildcard;

private:
    bool patternEqualsText();
    bool patternContainsWildcard();
    bool matchWithWildcards();
    bool startOrEndDoesNotMatch(const QStringList& parts);
    bool partsMatch(const QStringList& parts);
    int getMatchIndex(const QString& part, int startIndex);
    bool noMatchFound(int index);
    int calculateNewIndex(int matchIndex, int partLength);

    static const Qt::CaseSensitivity Sensitivity;
    const QString m_text;
    QString m_pattern;
};

#endif // KEEPASSX_WILDCARDMATCHER_H
